@@ -39,8 +39,6 @@ const char* stmt_name(StmtTag t) {
         "next",
         "break",
         "return",
-        "add",
-        "delete",
         "list",
         "bodylist",
         "<init>",
@@ -53,6 +51,7 @@ const char* stmt_name(StmtTag t) {
         "null",
         "assert",
         "extern",
+        "std-function",
     };
 
 #pragma GCC diagnostic push
@@ -903,49 +902,6 @@ TraversalCode SwitchStmt::Traverse(TraversalCallback* cb) const {
     HANDLE_TC_STMT_POST(tc);
 }
 
-AddDelStmt::AddDelStmt(StmtTag t, ExprPtr arg_e) : ExprStmt(t, std::move(arg_e)) {}
-
-bool AddDelStmt::IsPure() const { return false; }
-
-TraversalCode AddDelStmt::Traverse(TraversalCallback* cb) const {
-    TraversalCode tc = cb->PreStmt(this);
-    HANDLE_TC_STMT_PRE(tc);
-
-    // Argument is stored in base class's "e" field.
-    tc = e->Traverse(cb);
-    HANDLE_TC_STMT_PRE(tc);
-
-    tc = cb->PostStmt(this);
-    HANDLE_TC_STMT_POST(tc);
-}
-
-AddStmt::AddStmt(ExprPtr arg_e) : AddDelStmt(STMT_ADD, std::move(arg_e)) {
-    if ( ! e->CanAdd() )
-        Error("illegal add statement");
-}
-
-ValPtr AddStmt::Exec(Frame* f, StmtFlowType& flow) {
-    RegisterAccess();
-    flow = FLOW_NEXT;
-    e->Add(f);
-    return nullptr;
-}
-
-DelStmt::DelStmt(ExprPtr arg_e) : AddDelStmt(STMT_DELETE, std::move(arg_e)) {
-    if ( e->IsError() )
-        return;
-
-    if ( ! e->CanDel() )
-        Error("illegal delete statement");
-}
-
-ValPtr DelStmt::Exec(Frame* f, StmtFlowType& flow) {
-    RegisterAccess();
-    flow = FLOW_NEXT;
-    e->Delete(f);
-    return nullptr;
-}
-
 EventStmt::EventStmt(EventExprPtr arg_e) : ExprStmt(STMT_EVENT, arg_e), event_expr(std::move(arg_e)) {}
 
 ValPtr EventStmt::Exec(Frame* f, StmtFlowType& flow) {
@@ -1918,13 +1874,22 @@ void WhenInfo::Build(StmtPtr ws) {
 void WhenInfo::Instantiate(Frame* f) { Instantiate(lambda->Eval(f)); }
 
 void WhenInfo::Instantiate(ValPtr func) {
-    curr_lambda = with_location_of(make_intrusive<ConstExpr>(std::move(func)), cond);
+    curr_lambda = make_intrusive<ConstExpr>(std::move(func));
+    if ( cond )
+        curr_lambda->SetLocationInfo(cond->GetLocationInfo());
 }
 
-ExprPtr WhenInfo::Cond() { return with_location_of(make_intrusive<CallExpr>(curr_lambda, invoke_cond), cond); }
+ExprPtr WhenInfo::Cond() {
+    if ( cond )
+        return with_location_of(make_intrusive<CallExpr>(curr_lambda, invoke_cond), cond);
+    else
+        return make_intrusive<CallExpr>(curr_lambda, invoke_cond);
+}
 
 StmtPtr WhenInfo::WhenBody() {
-    auto invoke = with_location_of(make_intrusive<CallExpr>(curr_lambda, invoke_s), s);
+    auto invoke = make_intrusive<CallExpr>(curr_lambda, invoke_s);
+    if ( s )
+        invoke->SetLocationInfo(s->GetLocationInfo());
     return make_intrusive<ReturnStmt>(invoke, true);
 }
 
@@ -1954,11 +1919,14 @@ void WhenInfo::BuildInvokeElems() {
     invoke_s = make_intrusive<ListExpr>(two_const);
     invoke_timeout = make_intrusive<ListExpr>(three_const);
 
-    auto cl = cond->GetLocationInfo();
+    if ( cond ) {
+        // "cond" might not exist if we're constructing via -O gen-C++.
+        auto cl = cond->GetLocationInfo();
 
-    for ( const auto& e :
-          std::vector<ExprPtr>{one_const, two_const, three_const, invoke_cond, invoke_s, invoke_timeout} )
-        e->SetLocationInfo(cl);
+        for ( const auto& e :
+              std::vector<ExprPtr>{one_const, two_const, three_const, invoke_cond, invoke_s, invoke_timeout} )
+            e->SetLocationInfo(cl);
+    }
 }
 
 WhenStmt::WhenStmt(std::shared_ptr<WhenInfo> arg_wi) : Stmt(STMT_WHEN), wi(std::move(arg_wi)) {
@@ -2016,7 +1984,10 @@ void WhenStmt::StmtDescribe(ODesc* d) const {
             if ( c.IsDeepCopy() )
                 d->Add("copy ");
 
-            d->Add(c.Id()->Name());
+            if ( c.Id() )
+                d->Add(c.Id()->Name());
+            else
+                d->Add("<error>");
         }
         d->Add("]");
     }
@@ -2071,4 +2042,14 @@ TraversalCode WhenStmt::Traverse(TraversalCallback* cb) const {
     HANDLE_TC_STMT_POST(tc);
 }
 
+ValPtr StdFunctionStmt::Exec(Frame* f, StmtFlowType& flow) {
+    zeek::Args args = *f->GetFuncArgs();
+
+    // Set this to NEXT by default. The function can override that if it wants.
+    flow = FLOW_NEXT;
+    func(args, flow);
+
+    return nullptr;
+}
+
 } // namespace zeek::detail
